# Standard library imports
from datetime import datetime, timezone

# Third party imports
import jwt
import requests

# Project imports
from magma_auth.const import URL_VALIDATE_TOKEN


def validate_token(token: str, url: str = None) -> bool:
    """Validate token

    Args:
        token (str): token
        url (str, optional): url to validate

    Returns:
        bool: True if token is valid
    """
    headers = {"Authorization": "Bearer " + token}
    url = url or URL_VALIDATE_TOKEN

    try:
        response = requests.request("GET", url, headers=headers).json()
    except Exception as e:
        print(f"Error validating token: {e}")
        return False

    if "code" in response:
        if response["code"] == 419:
            return False

    return True


def decode_token(token: str = None) -> dict:
    """Decode token

    Args:
        token (str): Token to decode

    Returns:
        dict: Decoded token with keys: 'issued_at', 'expires_at', abd 'roles'
    """
    decoded = jwt.decode(token, options={"verify_signature": False})

    return {
        "issued_at": datetime.fromtimestamp(decoded["iat"], timezone.utc),
        "expired_at": datetime.fromtimestamp(decoded["exp"], timezone.utc),
        "roles": decoded["roles"] if "roles" in decoded.keys() else None,
    }



# Standard library imports
import base64
import json
import os

# Third party imports
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class Encryptor:
    def __init__(self):

        ssh_dir = os.path.join(os.path.expanduser("~"), ".ssh")
        os.makedirs(ssh_dir, exist_ok=True)

        self.private_key_path = os.path.join(ssh_dir, "private_key.pem")
        self.public_key_path = os.path.join(ssh_dir, "public_key.pem")

        self.private_key, self.public_key = self.load_or_generate_keys()
        self.AES_KEY = os.urandom(32)
        self._text = None
        self._encrypted_package = None

    @property
    def text(self):
        """Get unencrypted text."""
        return self._text

    @text.setter
    def text(self, value):
        """Set unencrypted text."""
        self._text = value

    @property
    def encrypted_package(self):
        return self._encrypted_package

    @encrypted_package.setter
    def encrypted_package(self, encrypted_package):
        self._encrypted_package = encrypted_package

    @staticmethod
    def generate_rsa_keys():
        """Generate RSA public and private key pair"""
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        public_key = private_key.public_key()
        return private_key, public_key

    def encrypt_with_aes(self):
        """Encrypt data using AES-GCM"""
        data = self.text
        iv = os.urandom(12)
        cipher = Cipher(algorithms.AES(self.AES_KEY), modes.GCM(iv))
        encryptor = cipher.encryptor()

        # Convert string to bytes if needed
        if isinstance(data, str):
            data = data.encode("utf-8")

        ciphertext = encryptor.update(data) + encryptor.finalize()

        return iv, ciphertext, encryptor.tag

    def decrypt_with_aes(self, iv, ciphertext, tag):
        """Decrypt data using AES-GCM"""
        cipher = Cipher(algorithms.AES(self.AES_KEY), modes.GCM(iv, tag))
        decryptor = cipher.decryptor()

        plaintext = decryptor.update(ciphertext) + decryptor.finalize()

        return plaintext.decode("utf-8")

    def hybrid_encrypt(self) -> str:
        """
        Hybrid encryption: Use AES for data, RSA for AES key
        Returns a dictionary with all encrypted components
        """
        # Encrypt the data with AES
        iv, ciphertext, tag = self.encrypt_with_aes()

        # Encrypt the AES key with RSA
        encrypted_aes_key = self.public_key.encrypt(
            self.AES_KEY,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

        # Package everything together
        encrypted_package = {
            "encrypted_aes_key": base64.b64encode(encrypted_aes_key).decode("utf-8"),
            "iv": base64.b64encode(iv).decode("utf-8"),
            "ciphertext": base64.b64encode(ciphertext).decode("utf-8"),
            "tag": base64.b64encode(tag).decode("utf-8"),
        }

        self.encrypted_package = encrypted_package

        # Return as JSON string for easy storage/transmission
        return json.dumps(encrypted_package)

    def hybrid_decrypt(self) -> str:
        """
        Hybrid decryption: Decrypt AES key with RSA, then decrypt data with AES
        """
        # Parse the JSON package
        encrypted_package = json.loads(json.dumps(self.encrypted_package))

        # Decode base64 components
        encrypted_aes_key = base64.b64decode(encrypted_package["encrypted_aes_key"])
        iv = base64.b64decode(encrypted_package["iv"])
        ciphertext = base64.b64decode(encrypted_package["ciphertext"])
        tag = base64.b64decode(encrypted_package["tag"])

        # Decrypt the AES key with RSA
        self.AES_KEY = self.private_key.decrypt(
            encrypted_aes_key,
            padding.OAEP(
                mgf=padding.MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None,
            ),
        )

        plaintext = self.decrypt_with_aes(iv, ciphertext, tag)

        return plaintext

    def save_keys(self, private_key, public_key):
        """Save keys to PEM files"""
        private_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        with open(self.private_key_path, "wb") as f:
            f.write(private_pem)

        # Save public key
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )
        with open(self.public_key_path, "wb") as f:
            f.write(public_pem)

    def load_keys(self):
        """Load keys from PEM files"""
        with open(self.private_key_path, "rb") as f:
            private_key = serialization.load_pem_private_key(f.read(), password=None)

        # Load public key
        with open(self.public_key_path, "rb") as f:
            public_key = serialization.load_pem_public_key(f.read())

        return private_key, public_key

    def load_or_generate_keys(self):
        if os.path.isfile(self.private_key_path) and os.path.isfile(
            self.public_key_path
        ):
            return self.load_keys()

        private_key, public_key = Encryptor.generate_rsa_keys()
        self.save_keys(private_key, public_key)
        return private_key, public_key



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from docutils.nodes import version

# Third party imports

# Project imports
from magma_auth.auth import Auth
from magma_auth.auth import AuthExternal
from magma_auth.utils import validate_token

__version__ = version("magma-auth")
__author__ = "Martanto"
__author_email__ = "martanto@live.com"
__license__ = "MIT"
__copyright__ = "Copyright (c) 2025, MAGMA Indonesia"
__url__ = "https://github.com/martanto/magma-auth"

auth = Auth()
auth_external = AuthExternal()

__all__ = [
    "__version__",
    "__author__",
    "__author_email__",
    "__license__",
    "__copyright__",
    "auth",
    "auth_external",
    "validate_token",
]



# Standard library imports
import json
import os
from datetime import datetime, timedelta, timezone
from typing import Self

# Third party imports
import requests
from magma_database import config
from requests import RequestException

# Project imports
from magma_auth.const import URL_LOGIN, URL_LOGIN_STAKEHOLDER
from magma_auth.encryptor import Encryptor
from magma_auth.utils import decode_token, validate_token

database_location = config["DATABASE_LOCATION"]
headers = {"Content-Type": "application/json"}


class Auth:
    def __init__(self):
        self.expired_at = None
        self.success = False
        self.encryptor = Encryptor()
        self._token = None
        self._token_file_location = os.path.join(database_location, ".token")
        self._credential_location = os.path.join(database_location, ".credential")

    @property
    def token(self) -> str | None:
        if self._token is not None:
            return self.encryptor.hybrid_decrypt()

        if os.path.exists(self._token_file_location):
            with open(self._token_file_location, "r") as f:
                self.encryptor.encrypted_package = json.loads(f.read())

            token = self.encryptor.hybrid_decrypt()
            decoded = decode_token(token)
            self.expired_at = decoded["expired_at"]

            # Remove expired token
            now = datetime.now(timezone.utc)
            if bool(now > self.expired_at):
                try:
                    os.remove(self._token_file_location)
                except Exception as e:
                    raise Exception(f"Error deleting {self._token_file_location}: {e}")

            self.success = True
            return token

        print(f"Token not found. Try login() first.")
        return None

    def login(
        self,
        username: str,
        password: str,
        ttl: int = 3,
        overwrite: bool = True,
        verbose: bool = False,
    ) -> bool:
        """Login using username and password of MAGMA Indonesia

        Args:
            username (str): username of MAGMA Indonesia
            password (str): password of MAGMA Indonesia
            ttl (int, optional): TTL of token as a day. Defaults to 3 day.
            overwrite (bool, optional): overwrite existing credentials. Defaults to True.
            verbose (bool, optional): verbose mode. Defaults to False.

        Returns:
            bool: True if login was successful

        Raises:
            RequestException: If authentication fails
        """
        payload = json.dumps({"username": username, "password": password, "ttl": ttl})

        try:
            response: dict = requests.request(
                "POST", URL_LOGIN, headers=headers, data=payload
            ).json()

            if not response["success"]:
                print(f"Wrong username or password.")
                return False

            self.success = self.save_token(response["token"], revalidate=False)
            self.expired_at = datetime.now() + timedelta(days=ttl)
            self.expired_at = self.expired_at.astimezone(timezone.utc)

            with open(self._token_file_location, "w") as f:
                f.write(json.dumps(self.encryptor.encrypted_package))

            # Save credential
            if overwrite:
                encryptor = Encryptor()
                encryptor.text = payload
                encrypted_payload = encryptor.hybrid_encrypt()

                with open(self._credential_location, "w") as f:
                    f.write(encrypted_payload)

            if verbose:
                print("Logged in successfully.")

            return True
        except RequestException as e:
            raise RequestException(f"Cannot login to MAGMA at the moment. \n{e}")

    def renew(
        self,
        username: str = "username",
        password: str = "password",
        ttl: int = 3,
        verbose: bool = False,
    ) -> bool:
        """Renew token using saved credential.
        Make sure you already have logged in once before.

        Args:
            username (str): Field username of MAGMA Indonesia
            password (str): Field password of MAGMA Indonesia
            ttl (int, optional): TTL of token as a day. Defaults to 3 day.
            verbose (bool, optional): verbose mode. Defaults to False.

        Returns:
            Self: self
        """
        if not os.path.exists(self._credential_location):
            print("Credential not found. Please login first.")
            return False

        try:
            encryptor = Encryptor()
            with open(self._credential_location, "r") as f:
                encryptor.encrypted_package = json.loads(f.read())

            payload = json.loads(encryptor.hybrid_decrypt())
            if self.login(
                payload[username],
                payload[password],
                ttl,
                overwrite=False,
                verbose=verbose,
            ):
                return True

            return False
        except RuntimeError as e:
            print(f"Failed to login using auto method. Try using login method. \n{e}")
            return False

    def save_token(self, token: str, revalidate: bool = True) -> bool:
        """Save token to file.

        Args:
            token (str): token to save
            revalidate (bool, optional): revalidate token. Defaults to True.

        Returns:
            bool: True if save was successful
        """
        token_is_valid = validate_token(token) if revalidate else True

        if token_is_valid:
            self.encryptor.text = token
            self._token: str = self.encryptor.hybrid_encrypt()
            with open(self._token_file_location, "w") as f:
                f.write(json.dumps(self.encryptor.encrypted_package))

            self.success = True
            decoded = decode_token(token)
            self.expired_at = decoded["expired_at"]

            return True

        return False


class AuthExternal(Auth):
    def __init__(self):
        super().__init__()
        self._token_file_location = os.path.join(database_location, ".x-token")
        self._credential_location = os.path.join(database_location, ".x-credential")

    def login(
        self,
        app_id: str,
        secret_key: str,
        overwrite: bool = True,
        verbose: bool = False,
        **kwargs,
    ) -> bool:
        """Login using username and password of MAGMA Indonesia

        Args:
            app_id (str): APP ID Stakeholder of MAGMA Indonesia
            secret_key (str): Secret Key of App ID of MAGMA Indonesia
            overwrite (bool, optional): overwrite existing credentials. Defaults to True.
            verbose (bool, optional): verbose mode. Defaults to False.
            kwargs (dict, optional): additional arguments passed to requests.request. Defaults to None.

        Returns:
            bool: True if login was successful

        Raises:
            RequestException: If authentication fails
        """
        payload = json.dumps({"app_id": app_id, "secret_key": secret_key})

        try:
            response: dict = requests.request(
                "POST", URL_LOGIN_STAKEHOLDER, headers=headers, data=payload
            ).json()

            if "message" in response.keys():
                print(f"Wrong app_id or secret_key.")
                return False

            self.success = self.save_token(response["token"], revalidate=False)
            self.expired_at = datetime.strptime(
                response["expired_at"],
                "%Y-%m-%d %H:%M:%S",
            )

            # Save credential
            if overwrite:
                encryptor = Encryptor()
                encryptor.text = payload
                encrypted_payload = encryptor.hybrid_encrypt()

                with open(self._credential_location, "w") as f:
                    f.write(encrypted_payload)

            if verbose:
                print("Logged in successfully.")

            return True
        except RequestException as e:
            raise RequestException(f"Cannot login to MAGMA at the moment. \n{e}")

    def renew(
        self,
        username: str = "app_id",
        password: str = "secret_key",
        verbose: bool = False,
        **kwargs,
    ) -> bool:
        """Renew token using saved credential.
        Make sure you already have logged in once before.

        Args:
            username (str): Field username of MAGMA Indonesia
            password (str): Field password of MAGMA Indonesia
            verbose (bool, optional): verbose mode. Defaults to False.
            kwargs (dict, optional): additional arguments passed to requests.request. Defaults to None.

        Returns:
            Self: self
        """
        if not os.path.exists(self._credential_location):
            print("Credential not found. Please login first.")
            return False

        try:
            encryptor = Encryptor()
            with open(self._credential_location, "r") as f:
                encryptor.encrypted_package = json.loads(f.read())

            payload = json.loads(encryptor.hybrid_decrypt())
            if self.login(
                payload[username],
                payload[password],
                overwrite=False,
                verbose=verbose,
            ):
                return True

            return False
        except RuntimeError as e:
            print(f"Failed to login using auto method. Try using login method. \n{e}")
            return False
